@@ -98,7 +98,10 @@ const char *errNames[kXR_ERRFENCE-kXR_ArgInvalid] =
                     "File attribute not found",   // kXR_AttrNotFound
                     "Operation requires TLS",     // kXR_TLSRequired
                     "No new servers for replica", // kXR_noReplicas
-                    "Authentication failed"       // kXR_AuthFailed
+                    "Authentication failed",      // kXR_AuthFailed
+                    "Request is not possible",    // kXR_Impossible
+                    "Conflicting request",        // kXR_Conflict
+                    "Too many errors"             // kXR_TooManyErrs
                    };
 
 const char *reqNames[kXR_REQFENCE-kXR_auth] =
