@@ -88,7 +88,7 @@
 #include "XrdTls/XrdTls.hh"
 #include "XrdTls/XrdTlsContext.hh"
 
-#ifdef __linux__
+#if defined(__linux__) || defined(__GNU__)
 #include <netinet/tcp.h>
 #endif
 #ifdef __APPLE__
@@ -168,13 +168,44 @@ XrdConfigProt  *Next;
 char           *proname;
 char           *libpath;
 char           *parms;
-int             port;
-bool            dotls;
+
+int             numP;
+union {int      port;
+       int      portVec[XrdProtLoad::PortoMax];
+      };
+union {bool     dotls;
+       bool     tlsVec[XrdProtLoad::PortoMax];
+      };
+
+bool            AddPort(int pnum, bool isTLS)
+                       {for (int i = 0; i < numP; i++)
+                            if (pnum == portVec[i])
+                               {tlsVec[i] = isTLS; return true;}
+                        if (numP >= (XrdProtLoad::PortoMax)) return false;
+                        portVec[numP] = pnum; tlsVec[numP] = isTLS;
+                        numP++;
+                        return true;
+                       }
+
+void             Reset(char *ln, char *pp, int np=-1, bool to=false)
+                      {if (libpath) free(libpath);
+                       libpath = ln;
+                       if (parms)   free(parms);
+                       parms = pp;
+                       memset(portVec, 0, sizeof(portVec));
+                       port = np;
+                       memset(tlsVec, 0, sizeof(tlsVec));
+                       dotls = to;
+                       numP = 1;
+                      }
 
                 XrdConfigProt(char *pn, char *ln, char *pp, int np=-1,
                               bool to=false)
-                    : Next(0), proname(pn), libpath(ln), parms(pp), port(np),
-                      dotls(to) {}
+                    : Next(0), proname(pn), libpath(ln), parms(pp), numP(1)
+                    {memset(portVec, 0, sizeof(portVec)); port = np;
+                     memset(tlsVec, 0, sizeof(tlsVec)); dotls = to;
+                    }
+
                ~XrdConfigProt()
                     {free(proname);
                      if (libpath) free(libpath);
@@ -212,7 +243,7 @@ XrdConfig::XrdConfig()
 //
    PortTCP  = -1;
    PortUDP  = -1;
-   PortTLS  = 0;
+   PortTLS  = -1;
    ConfigFN = 0;
    tmoInfo  = 0;
    myInsName= 0;
@@ -238,11 +269,9 @@ XrdConfig::XrdConfig()
    tlsOpts    = 9ULL | XrdTlsContext::servr | XrdTlsContext::logVF;
    tlsNoVer   = false;
    tlsNoCAD   = true;
-   NetTCPlep  = -1;
    NetADM     = 0;
    coreV      = 1;
    Specs      = 0;
-   memset(NetTCP, 0, sizeof(NetTCP));
 
    Firstcp = Lastcp = 0;
 
@@ -306,6 +335,12 @@ int XrdConfig::Configure(int argc, char **argv)
    int   myArgc = 1, urArgc = argc, i;
    bool noV6, ipV4 = false, ipV6 = false, rootChk = true, optbg = false;
 
+// Reconstruct the command line so we can put it in the log
+//
+   XrdOucString CmdLine(argv[0]);
+   for (int k = 1; k < argc; k++)
+       {CmdLine += ' '; CmdLine += argv[k];}
+
 // Obtain the program name we will be using
 //
     retc = strlen(argv[0]);
@@ -411,7 +446,8 @@ int XrdConfig::Configure(int argc, char **argv)
                  break;
        case 'p': if ((clPort = yport(&Log, "tcp", optarg)) < 0) Usage(1);
                  break;
-       case 'P': dfltProt = optarg;
+       case 'P': if (dfltProt) free(dfltProt);
+                 dfltProt = strdup(optarg);
                  break;
        case 'R': if (!(getUG(optarg, myUid, myGid))) Usage(1);
                  rootChk = false;
@@ -562,6 +598,7 @@ int XrdConfig::Configure(int argc, char **argv)
        LogInfo.iName  = myInsName;
        LogInfo.cfgFn  = ConfigFN;
        if (!XrdOucLogging::configLog(Log, LogInfo)) _exit(16);
+       Log.logger()->AddMsg(CmdLine.c_str());
        Log.logger()->AddMsg(XrdBANNER);
       }
 
@@ -586,6 +623,7 @@ int XrdConfig::Configure(int argc, char **argv)
    retc = strlen(buff);
    XrdSysUtils::FmtUname(buff+retc, sizeof(buff)-retc);
    Log.Say(0, buff);
+   Log.Say(0, CmdLine.c_str());
    Log.Say(XrdBANNER);
 
 // Verify that we have a real name. We've had problems with people setting up
@@ -651,7 +689,7 @@ int XrdConfig::Configure(int argc, char **argv)
       }
 
 // If there is TLS port verify that it can be used. We ignore this if we
-// will fal anyway so as to not issue confusing messages.
+// will fail anyway so as to not issue confusing messages.
 //
    if (!NoGo)
       {if (PortTLS > 0 && !XrdGlobal::tlsCtx)
@@ -659,7 +697,7 @@ int XrdConfig::Configure(int argc, char **argv)
            PortTLS = -1;
           } else {
            ProtInfo.tlsCtx  = XrdGlobal::tlsCtx;
-           ProtInfo.tlsPort = PortTLS;
+           ProtInfo.tlsPort = (PortTLS > 0 ? PortTLS : 0);
          }
       }
 
@@ -793,12 +831,12 @@ int XrdConfig::ConfigXeq(char *var, XrdOucStream &Config, XrdSysError *eDest)
   
 int XrdConfig::ASocket(const char *path, const char *fname, mode_t mode)
 {
-   struct sockaddr_un unix;
+   struct sockaddr_un unixvar;
    int  plen = strlen(path), flen = strlen(fname);
 
 // Make sure we can fit everything in our buffer
 //
-   if ((plen + flen + 3) > (int)sizeof(unix.sun_path))
+   if ((plen + flen + 3) > (int)sizeof(unixvar.sun_path))
       {Log.Emsg("Config", "admin path", path, "too long");
        return 1;
       }
@@ -869,6 +907,44 @@ int XrdConfig::ConfigProc()
    return NoGo;
 }
 
+/******************************************************************************/
+/*                                g e t N e t                                 */
+/******************************************************************************/
+
+XrdInet *XrdConfig::getNet(int port, bool isTLS)
+{
+   int the_Opts, the_Blen;
+
+// Try to find an existing network for this port
+//
+   for (int i = 0; i < (int)NetTCP.size(); i++)
+       if (port == NetTCP[i]->Port()) return NetTCP[i];
+
+// Create a new network for this port
+//
+   XrdInet *newNet = new XrdInet(&Log, Police);
+   NetTCP.push_back(newNet);
+
+// Set options
+//
+   if (isTLS)
+      {the_Opts = TLS_Opts; the_Blen = TLS_Blen;
+      } else {
+       the_Opts = Net_Opts; the_Blen = Net_Blen;
+      }
+   if (the_Opts || the_Blen) newNet->setDefaults(the_Opts, the_Blen);
+
+// Set the domain if we have one
+//
+   if (myDomain) newNet->setDomain(myDomain);
+
+// Attempt to bind to this socket.
+//
+   if (newNet->BindSD(port, "tcp") == 0) return newNet;
+   delete newNet;
+   return 0;
+}
+  
 /******************************************************************************/
 /*                                 g e t U G                                  */
 /******************************************************************************/
@@ -1143,7 +1219,7 @@ int XrdConfig::setFDL()
 
 // The scheduler will have already set the thread limit. We just report it
 //
-#if defined(__linux__) && defined(RLIMIT_NPROC)
+#if ( defined(__linux__) || defined(__GNU__) || (defined(__FreeBSD_kernel__) && defined(__GLIBC__)) ) && defined(RLIMIT_NPROC)
 
 // Obtain the actual limit now (Scheduler construction sets this to rlim_max)
 //
@@ -1169,7 +1245,7 @@ int XrdConfig::setFDL()
 int XrdConfig::Setup(char *dfltp, char *libProt)
 {
    XrdConfigProt *cp;
-   int i, xport, wsz, arbNet, the_Opts, the_Blen;
+   int xport, protNum = 0;
 
 // Establish the FD limit
 //
@@ -1177,7 +1253,7 @@ int XrdConfig::Setup(char *dfltp, char *libProt)
 
 // Special handling for Linux sendfile()
 //
-#if defined(__linux__) && defined(TCP_CORK)
+#if ( defined(__linux__) || defined(__GNU__) ) && defined(TCP_CORK)
 {  int sokFD, setON = 1;
    if ((sokFD = socket(PF_INET, SOCK_STREAM, 0)) >= 0)
       {setsockopt(sokFD, XrdNetUtils::ProtoID("tcp"), TCP_NODELAY,
@@ -1213,32 +1289,26 @@ int XrdConfig::Setup(char *dfltp, char *libProt)
           else PortTCP = -1;
       }
 
-// If there is a TLS port then add the default protocol using tls
-//
-   if (PortTLS > 0)
-      {char pBuff[80];
-       snprintf(pBuff, sizeof(pBuff), "%ss", dfltp);
-       cp = new XrdConfigProt(strdup(pBuff), libProt, 0, PortTLS, true);
-       cp->Next = Firstcp;
-       Firstcp = cp;
-      }
-
 // We now go through all of the protocols and get each respective port number.
 //
    cp = Firstcp;
    while(cp)
-        {if (cp->dotls)
-            {if (!tlsCtx)
-                {Log.Emsg("Config", "protocol", cp->proname,
-                          "requires TLS but TLS is not configured!");
-                 return 1;
+        {if (!tlsCtx)
+            for (int i = 0; i < cp->numP; i++)
+                {if (cp->tlsVec[i])
+                    {Log.Emsg("Config", "protocol", cp->proname,
+                              "configured with a TLS-only port "
+                              "but TLS is not configured!");
+                     return 1;
+                    }
                 }
-             xport = PortTLS;
-            } else xport = PortTCP;
+         xport = (cp->dotls ? PortTLS : PortTCP);
          ProtInfo.Port = (cp->port < 0 ? xport : cp->port);
          XrdOucEnv::Export("XRDPORT", ProtInfo.Port);
-         if ((cp->port = XrdProtLoad::Port(cp->libpath, cp->proname,
-                                           cp->parms, &ProtInfo)) < 0) return 1;
+         cp->port = XrdProtLoad::Port(cp->libpath,cp->proname,cp->parms,&ProtInfo);
+         if (cp->port < 0) return 1;
+         for (int i = 1; i < cp->numP; i++)
+             if (cp->port == cp->portVec[i]) cp->portVec[i] = -1;
          cp = cp->Next;
         }
 
@@ -1249,46 +1319,55 @@ int XrdConfig::Setup(char *dfltp, char *libProt)
                                  ProtInfo.myName, Firstcp->port,
                                  ProtInfo.myInst, ProtInfo.myProg, mySitName);
 
+// If the base protocol is xroot, then save the base port number so we can
+// extend the port to the http protocol should it have been loaded. That way
+// redirects via xroot will also work for http.
+//
+   xport = (strcmp("xroot", Firstcp->proname) ? 0 : Firstcp->port);
+
 // Load the protocols. For each new protocol port number, create a new
 // network object to handle the port dependent communications part. All
-// port issues will have been resolved at this point.
+// port issues will have been resolved at this point. Note that we need
+// to set default network object from the first protocol before loading
+// any protocol in case one of them starts using the default network.
 //
-   arbNet = XrdProtLoad::ProtoMax;
+   XrdInet *arbNet = 0, *theNet;
    while((cp = Firstcp))
-        {if (!(cp->port)) i = arbNet;
-            else for (i = 0; i < XrdProtLoad::ProtoMax && NetTCP[i]; i++)
-                     {if (cp->port == NetTCP[i]->Port()) break;}
-         if (i >= XrdProtLoad::ProtoMax || !NetTCP[i])
-            {NetTCP[++NetTCPlep] = new XrdInet(&Log, Police);
-             if (cp->dotls)
-                {the_Opts = TLS_Opts; the_Blen = TLS_Blen;
-                } else {
-                 the_Opts = Net_Opts; the_Blen = Net_Blen;
-                }
-             if (the_Opts || the_Blen)
-                NetTCP[NetTCPlep]->setDefaults(the_Opts, the_Blen);
-             if (myDomain) NetTCP[NetTCPlep]->setDomain(myDomain);
-             if (NetTCP[NetTCPlep]->BindSD(cp->port, "tcp")) return 1;
-             ProtInfo.Port   = NetTCP[NetTCPlep]->Port();
-             ProtInfo.NetTCP = NetTCP[NetTCPlep];
-             wsz             = NetTCP[NetTCPlep]->WSize();
-             ProtInfo.WSize  = (wsz < the_Blen || !the_Blen ? wsz : the_Blen);
-             TRACE(NET, cp->proname <<" port " <<ProtInfo.Port <<" wsz="
-                        <<ProtInfo.WSize <<" (" <<wsz <<')');
-             if (!(cp->port)) arbNet = NetTCPlep;
-             if (!NetTCPlep) XrdNetTCP = NetTCP[0];
-             XrdOucEnv::Export("XRDPORT", ProtInfo.Port);
-            }
-         if (!XrdProtLoad::Load(cp->libpath,cp->proname,cp->parms,
-                               &ProtInfo, cp->dotls)) return 1;
-         Firstcp = cp->Next; delete cp;
-        }
+         {for (int i = 0; i < cp->numP; i++)
+             {if (cp->portVec[i] < 0) continue;
+              if (!(cp->portVec[i]) && arbNet) theNet = arbNet;
+                 else {theNet = getNet(cp->portVec[i], cp->tlsVec[i]);
+                       if (!theNet) return 1;
+                       if (!(cp->portVec[i])) arbNet = theNet;
+                      }
+              if (i == 0) XrdNetTCP = theNet; // Avoid race condition!!!
+              ProtInfo.Port   = theNet->Port();
+              ProtInfo.NetTCP = theNet;
+              ProtInfo.WSize  = theNet->WSize();
+              TRACE(NET, cp->proname <<':' <<ProtInfo.Port <<" wsz="
+                         <<ProtInfo.WSize);
+
+              if (i) XrdProtLoad::Port(protNum, ProtInfo.Port, cp->tlsVec[i]);
+                 else {XrdOucEnv::Export("XRDPORT", ProtInfo.Port);
+                       protNum = XrdProtLoad::Load(cp->libpath, cp->proname,
+                                                   cp->parms, &ProtInfo,
+                                                   cp->dotls);
+                       if (!protNum) return 1;
+                      }
+             }
+          if (!strcmp("http", cp->proname) && xport)
+             {for (int i = 0; i < cp->numP; i++)
+                  {if (cp->portVec[i] == xport) {xport = 0; break;}}
+              if (xport) XrdProtLoad::Port(protNum, xport, false);
+             }
+          Firstcp = cp->Next; delete cp;
+         }
 
 // Leave the env port number to be the first used port number. This may
-// or may not be the same as the default port number.
+// or may not be the same as the default port number. This corresponds to
+// the default network object.
 //
-   ProtInfo.Port = NetTCP[0]->Port();
-   PortTCP = ProtInfo.Port;
+   PortTCP = ProtInfo.Port = XrdNetTCP->Port();
    XrdOucEnv::Export("XRDPORT", PortTCP);
 
 // Now check if we have to setup automatic reporting
@@ -1681,7 +1760,6 @@ int XrdConfig::xnet(XrdSysError *eDest, XrdOucStream &Config)
         {if (V_blen >= 0) TLS_Blen = V_blen;
          if (V_keep >= 0) TLS_Opts = (V_keep  ? XRDNET_KEEPALIVE : 0);
          TLS_Opts |= (V_nodnr ? XRDNET_NORLKUP   : 0) | XRDNET_USETLS;
-         if (!PortTLS) PortTLS = -1;
         } else {
          if (V_blen >= 0) Net_Blen = V_blen;
          if (V_keep >= 0) Net_Opts = (V_keep  ? XRDNET_KEEPALIVE : 0);
@@ -1843,8 +1921,9 @@ int XrdConfig::yport(XrdSysError *eDest, const char *ptype, const char *val)
 
 /* Function: xprot
 
-   Purpose:  To parse the directive: protocol [tls] <name>[:<port>] <loc> [<parm>]
+   Purpose:  To parse the directive: protocol [tls] <name>[:<port>] <args>
 
+             <args> {+port | <loc> [<parm>]}
              tls    The protocol requires tls.
              <name> The name of the protocol (e.g., rootd)
              <port> Port binding for the protocol, if not the default.
@@ -1857,8 +1936,8 @@ int XrdConfig::yport(XrdSysError *eDest, const char *ptype, const char *val)
 int XrdConfig::xprot(XrdSysError *eDest, XrdOucStream &Config)
 {
     XrdConfigProt *cpp;
-    char *val, *parms, *lib, proname[64], buff[1024];
-    int vlen, bleft = sizeof(buff), portnum = -1;
+    char *val, *parms, *lib, proname[64], buff[2048];
+    int portnum = -1;
     bool dotls = false;
 
     do {if (!(val = Config.GetWord()))
@@ -1871,49 +1950,67 @@ int XrdConfig::xprot(XrdSysError *eDest, XrdOucStream &Config)
        {eDest->Emsg("Config", "protocol name is too long"); return 1;}
     strcpy(proname, val);
 
-    if (!(val = Config.GetWord()))
-       {eDest->Emsg("Config", "protocol library not specified"); return 1;}
-    if (strcmp("*", val)) lib = strdup(val);
-       else lib = 0;
-
-    parms = buff;
-    while((val = Config.GetWord()))
-         {vlen = strlen(val); bleft -= (vlen+1);
-          if (bleft <= 0)
-             {eDest->Emsg("Config", "Too many parms for protocol", proname);
-              return 1;
-             }
-          *parms = ' '; parms++; strcpy(parms, val); parms += vlen;
-         }
-    if (parms != buff) parms = strdup(buff+1);
-       else parms = 0;
-
     if ((val = index(proname, ':')))
        {if ((portnum = yport(&Log, "tcp", val+1)) < 0) return 1;
            else *val = '\0';
-       } else {
-        if (dotls && !PortTLS) PortTLS = -1;
        }
 
+    if (!(val = Config.GetWord()))
+       {eDest->Emsg("Config", "protocol library not specified"); return 1;}
+    if (!strcmp("*", val)) lib = 0;
+       else if (*val == '+')
+               {if (strcmp(val, "+port"))
+                   {eDest->Emsg("Config","invalid library specification -",val);
+                    return 1;
+                   }
+                if ((cpp = Firstcp))
+                   do {if (!strcmp(proname, cpp->proname))
+                          {if (cpp->AddPort(portnum, dotls)) return 0;
+                           eDest->Emsg("Config", "port add limit exceeded!");
+                           return 1;
+                          }
+                      } while((cpp = cpp->Next));
+                eDest->Emsg("Config","protocol",proname,"not previously defined!");
+                return 1;
+               }
+               else lib = strdup(val);
+
+// If no library was specified then this is a default protocol. We must make sure
+// sure it is consistent with whatever default we have.
+//
+   if (!lib && Firstcp && strcmp(proname, Firstcp->proname))
+      {char eBuff[512];
+       snprintf(eBuff, sizeof(eBuff), "the %s protocol is '%s' not '%s'; "
+                "assuming you meant '%s'",
+                (Firstcp->libpath ? "assigned" : "builtin"),
+                 Firstcp->proname, proname, Firstcp->proname);
+       eDest->Say("Config warning: ", eBuff, " but please correct "
+                                      "the following directive!");
+       snprintf(proname, sizeof(proname), "%s", Firstcp->proname);
+      }
+
+    *buff = 0;
+    if (!Config.GetRest(buff, sizeof(buff)))
+       {eDest->Emsg("Config", "Too many parms for protocol", proname);
+        return 1;
+       }
+    parms = (*buff ? strdup(buff) : 0);
+
     if ((cpp = Firstcp))
        do {if (!strcmp(proname, cpp->proname))
-              {if (cpp->libpath) free(cpp->libpath);
-               if (cpp->parms)   free(cpp->parms);
-               cpp->libpath = lib;
-               cpp->parms   = parms;
-               cpp->port    = portnum;
-               cpp->dotls   = dotls;
+              {cpp->Reset(lib, parms, portnum, dotls);
                return 0;
               }
           } while((cpp = cpp->Next));
 
-//  if (lib)
-       {cpp = new XrdConfigProt(strdup(proname), lib, parms, portnum, dotls);
-        if (Lastcp) Lastcp->Next = cpp;
-           else    Firstcp = cpp;
-        Lastcp = cpp;
-       }
-
+    cpp = new XrdConfigProt(strdup(proname), lib, parms, portnum, dotls);
+    if (!lib) {cpp->Next = Firstcp; Firstcp = cpp;
+               if (!Lastcp) Lastcp = cpp;
+              }
+       else   {if (Lastcp) Lastcp->Next = cpp;
+                  else     Firstcp = cpp;
+               Lastcp = cpp;
+              }
     return 0;
 }
 
@@ -2022,7 +2119,8 @@ int XrdConfig::xrep(XrdSysError *eDest, XrdOucStream &Config)
 
 // All done
 //
-   if (!(repOpts & XRD_STATS_ALL)) repOpts = XRD_STATS_ALL & ~XRD_STATS_INFO;
+   if (!(repOpts & XRD_STATS_ALL))
+      repOpts = char(XRD_STATS_ALL & ~XRD_STATS_INFO);
    return 0;
 }
 
