@@ -109,9 +109,9 @@ namespace XrdCl
   struct StatInfoImpl
   {
       StatInfoImpl() : pSize( 0 ), pFlags( 0 ), pModifyTime( 0 ),
-                       pChangeTime( 0 ), pAccessTime( 0 ), pExtended( false )
+                       pChangeTime( 0 ), pAccessTime( 0 ),
+                       pExtended( false ), pHasCksum( false )
       {
-
       }
 
       StatInfoImpl( const StatInfoImpl & pimpl ) : pId( pimpl.pId ),
@@ -123,7 +123,8 @@ namespace XrdCl
                                                    pMode( pimpl.pMode ),
                                                    pOwner( pimpl.pOwner ),
                                                    pGroup( pimpl.pGroup ),
-                                                   pExtended( pimpl.pExtended )
+                                                   pExtended( pimpl.pExtended ),
+                                                   pHasCksum( pimpl.pHasCksum )
       {
       }
 
@@ -165,7 +166,7 @@ namespace XrdCl
           return false;
         }
 
-        if( chunks.size() == 9 )
+        if( chunks.size() >= 9 )
         {
           pChangeTime = ::strtoll( chunks[4].c_str(), &result, 0 );
           if( *result != 0 )
@@ -191,6 +192,16 @@ namespace XrdCl
           pExtended = true;
         }
 
+        // after the extended stat information, we might have the checksum
+        if( chunks.size() >= 10 )
+        {
+          if( ( chunks[9] == "[" ) && ( chunks[11] == "]" ) )
+          {
+            pHasCksum = true;
+            pCksum     = chunks[10];
+          }
+        }
+
         return true;
       }
 
@@ -205,6 +216,8 @@ namespace XrdCl
       std::string pGroup;
 
       bool        pExtended;
+      bool        pHasCksum;
+      std::string pCksum;
   };
 
   //----------------------------------------------------------------------------
@@ -263,6 +276,14 @@ namespace XrdCl
     return pImpl->pSize;
   }
 
+  //------------------------------------------------------------------------
+  //! Set size
+  //------------------------------------------------------------------------
+  void StatInfo::SetSize( uint64_t size )
+  {
+    pImpl->pSize = size;
+  }
+
   //------------------------------------------------------------------------
   //! Get flags
   //------------------------------------------------------------------------
@@ -382,6 +403,14 @@ namespace XrdCl
     return pImpl->pGroup;
   }
 
+  //------------------------------------------------------------------------
+  //! Get checksum
+  //------------------------------------------------------------------------
+  const std::string& StatInfo::GetChecksum() const
+  {
+    return pImpl->pCksum;
+  }
+
   //------------------------------------------------------------------------
   //! Parse server response and fill up the object
   //------------------------------------------------------------------------
@@ -390,6 +419,14 @@ namespace XrdCl
     return pImpl->pExtended;
   }
 
+  //------------------------------------------------------------------------
+  //! Has checksum
+  //------------------------------------------------------------------------
+  bool StatInfo::HasChecksum() const
+  {
+    return pImpl->pHasCksum;
+  }
+
   //----------------------------------------------------------------------------
   // StatInfo constructor
   //----------------------------------------------------------------------------
@@ -553,4 +590,137 @@ namespace XrdCl
     std::string dat = data;
     return !dat.compare( 0, dStatPrefix.size(), dStatPrefix );
   }
+
+  struct PageInfoImpl
+  {
+    PageInfoImpl( uint64_t offset = 0, uint32_t length = 0, void *buffer = 0,
+                  std::vector<uint32_t> &&cksums = std::vector<uint32_t>() ) :
+      offset( offset ),
+      length( length ),
+      buffer( buffer ),
+      cksums( std::move( cksums ) )
+    {
+    }
+
+    PageInfoImpl( PageInfoImpl &&pginf ) : offset( pginf.offset ),
+                                           length( pginf.length ),
+                                           buffer( pginf.buffer ),
+                                           cksums( std::move( pginf.cksums ) )
+    {
+    }
+
+    uint64_t               offset; //> offset in the file
+    uint32_t               length; //> length of the data read
+    void                  *buffer; //> buffer with the read data
+    std::vector<uint32_t>  cksums; //> a vector of crc32c checksums
+  };
+
+  //----------------------------------------------------------------------------
+  // Default constructor
+  //----------------------------------------------------------------------------
+  PageInfo::PageInfo( uint64_t offset, uint32_t length, void *buffer,
+                      std::vector<uint32_t> &&cksums ) :
+    pImpl( new PageInfoImpl( offset, length, buffer, std::move( cksums ) ) )
+  {
+  }
+
+  //----------------------------------------------------------------------------
+  // Move constructor
+  //----------------------------------------------------------------------------
+  PageInfo::PageInfo( PageInfo &&pginf ) : pImpl( std::move( pginf.pImpl ) )
+  {
+  }
+
+  //----------------------------------------------------------------------------
+  // Destructor
+  //----------------------------------------------------------------------------
+  PageInfo::~PageInfo()
+  {
+  }
+
+  //----------------------------------------------------------------------------
+  // Get the offset
+  //----------------------------------------------------------------------------
+  uint64_t PageInfo::GetOffset() const
+  {
+    return pImpl->offset;
+  }
+
+  //----------------------------------------------------------------------------
+  // Get the data length
+  //----------------------------------------------------------------------------
+  uint32_t PageInfo::GetLength() const
+  {
+    return pImpl->length;
+  }
+
+  //----------------------------------------------------------------------------
+  // Get the buffer
+  //----------------------------------------------------------------------------
+  void* PageInfo::GetBuffer()
+  {
+    return pImpl->buffer;
+  }
+
+  //----------------------------------------------------------------------------
+  // Get the buffer
+  //----------------------------------------------------------------------------
+  std::vector<uint32_t>& PageInfo::GetCksums()
+  {
+    return pImpl->cksums;
+  }
+
+  //------------------------------------------------------------------------
+  // Factory function for generating handler objects from lambdas
+  //------------------------------------------------------------------------
+  ResponseHandler* ResponseHandler::Wrap( std::function<void(XRootDStatus&, AnyObject&)> func )
+  {
+    struct FuncHandler : public ResponseHandler
+    {
+      FuncHandler( std::function<void(XRootDStatus&, AnyObject&)> func ) : func( std::move( func ) )
+      {
+      }
+
+      void HandleResponse( XRootDStatus *status, AnyObject *response )
+      {
+        // make sure the arguments will be released
+        std::unique_ptr<XRootDStatus> stptr( status );
+        std::unique_ptr<AnyObject> rspptr( response );
+        // if there is no response provide a null reference placeholder
+        static AnyObject nullref;
+        if( response == nullptr )
+          response = &nullref;
+        // call the user completion handler
+        func( *status, *response );
+        // deallocate the wrapper
+        delete this;
+      }
+
+      std::function<void(XRootDStatus&, AnyObject&)> func;
+    };
+
+    return new FuncHandler( func );
+  }
+
+  ResponseHandler* ResponseHandler::Wrap( std::function<void(XRootDStatus*, AnyObject*)> func )
+  {
+    struct FuncHandler : public ResponseHandler
+    {
+      FuncHandler( std::function<void(XRootDStatus*, AnyObject*)> func ) : func( std::move( func ) )
+      {
+      }
+
+      void HandleResponse( XRootDStatus *status, AnyObject *response )
+      {
+        // call the user completion handler
+        func( status, response );
+        // deallocate the wrapper
+        delete this;
+      }
+
+      std::function<void(XRootDStatus*, AnyObject*)> func;
+    };
+
+    return new FuncHandler( func );
+  }
 }
