@@ -603,6 +603,16 @@ int XrdHttpProtocol::Process(XrdLink *lp) // We ignore the argument here
 
     if (!CurrentReq.headerok) {
       TRACEI(REQ, " rc:" << rc << "Header not yet complete.");
+      
+      // Here a subtle error condition. IF we failed reading a line AND the buffer
+      // has a reasonable amount of data available THEN we consider the header
+      // as corrupted and shutdown the client
+      if ((rc <= 0) && (BuffUsed() >= 16384)) {
+        TRACEI(ALL, "Corrupted header detected, or line too long. Disconnecting client.");
+        return -1;
+      }
+        
+        
       if (CurrentReq.reqstate > 0)
         CurrentReq.reqstate--;
       // Waiting for more data
@@ -736,6 +746,7 @@ int XrdHttpProtocol::Process(XrdLink *lp) // We ignore the argument here
         nfo = CurrentReq.opaque->Get("xrdhttphost");
         if (nfo) {
           TRACEI(DEBUG, " Setting host: " << nfo);
+          if (SecEntity.host) free(SecEntity.host);
           SecEntity.host = unquote(nfo);
           TRACEI(REQ, " Setting host: " << SecEntity.host);
         }
@@ -1300,8 +1311,8 @@ int XrdHttpProtocol::getDataOneShot(int blen, bool wait) {
     }
 
     if (rlen < 0) {
-      Link->setEtext("link timeout");
-      return 1;
+      Link->setEtext("link timeout or other error");
+      return -1;
     }
   }
 
@@ -2789,6 +2800,7 @@ int XrdHttpProtocol::LoadExtHandler(std::vector<extHInfo> &hiVec,
   // Add the pointer to the cadir and the cakey to the environment.
   //
   if (sslcadir) myEnv.Put("http.cadir", sslcadir);
+  if (sslcafile) myEnv.Put("http.cafile", sslcafile);
   if (sslcert)  myEnv.Put("http.cert",  sslcert);
   if (sslkey)   myEnv.Put("http.key"  , sslkey);
 
