@@ -34,6 +34,9 @@
 #include <set>
 #include <stdio.h>
 #include <string>
+#include <unistd.h>
+#include <sys/stat.h>
+#include <sys/types.h>
 
 #include "XrdCl/XrdClDefaultEnv.hh"
 #include "XrdCl/XrdClJobManager.hh"
@@ -83,6 +86,7 @@ extern XrdCl::DirListFlags::Flags dlFlag;
 extern bool                       oidsOK;
 extern bool                       p2lSRC;
 extern bool                       p2lSGI;
+extern bool                       autoPGRD;
 };
   
 /******************************************************************************/
@@ -318,6 +322,54 @@ void XrdPosixConfig::initEnv(XrdOucEnv &theEnv, const char *vName, long long &De
           }
       }
 }
+
+/******************************************************************************/
+/*                              i n i t S t a t                               */
+/******************************************************************************/
+
+void XrdPosixConfig::initStat(struct stat *buf)
+{
+   static int initStat = 0;
+   static dev_t st_rdev;
+   static dev_t st_dev;
+   static uid_t myUID = getuid();
+   static gid_t myGID = getgid();
+
+// Initialize the xdev fields. This cannot be done in the constructor because
+// we may not yet have resolved the C-library symbols.
+//
+   if (!initStat) {initStat = 1; initXdev(st_dev, st_rdev);}
+   memset(buf, 0, sizeof(struct stat));
+
+// Preset common fields
+//
+   buf->st_blksize= 64*1024;
+   buf->st_dev    = st_dev;
+   buf->st_rdev   = st_rdev;
+   buf->st_nlink  = 1;
+   buf->st_uid    = myUID;
+   buf->st_gid    = myGID;
+}
+  
+/******************************************************************************/
+/*                              i n i t X d e v                               */
+/******************************************************************************/
+  
+void XrdPosixConfig::initXdev(dev_t &st_dev, dev_t &st_rdev)
+{
+   static dev_t tDev, trDev;
+   static bool aOK = false;
+   struct stat buf;
+
+// Get the device id for /tmp used by stat()
+//
+   if (aOK) {st_dev = tDev; st_rdev = trDev;}
+      else if (stat("/tmp", &buf)) {st_dev = 0; st_rdev = 0;}
+              else {st_dev  = tDev  = buf.st_dev;
+                    st_rdev = trDev = buf.st_rdev;
+                    aOK = true;
+                   }
+}
   
 /******************************************************************************/
 /*                                O p e n F C                                 */
@@ -349,6 +401,7 @@ bool XrdPosixConfig::OpenFC(const char *path, int oflag, mode_t mode,
 bool XrdPosixConfig::SetConfig(XrdOucPsx &parms)
 {
    XrdOucTList *tP;
+   const char *val;
 
 // Set log routing
 //
@@ -400,6 +453,15 @@ bool XrdPosixConfig::SetConfig(XrdOucPsx &parms)
        XrdPosixGlobals::ddInterval = (parms.cioWait  < 10 ? 10 : parms.cioWait);
       }
 
+// Set auto conversion of read to pgread
+//
+   if (parms.theCache && parms.theEnv && (val = parms.theEnv->Get("psx.CSNet")))
+      {if (*val == '1' || *val == '2')
+          {XrdPosixGlobals::autoPGRD = true;
+           if (*val == '2') SetEnv("WantTlsOnNoPgrw", 1);
+          }
+      }
+
 // Handle the caching options (library or builin memory).
 // TODO: Make the memory cache a library plugin as well.
 //
