@@ -26,11 +26,18 @@ using namespace TPC;
 uint64_t TPCHandler::m_monid{0};
 int TPCHandler::m_marker_period = 5;
 size_t TPCHandler::m_block_size = 16*1024*1024;
+size_t TPCHandler::m_small_block_size = 1*1024*1024;
 XrdSysMutex TPCHandler::m_monid_mutex;
 
 XrdVERSIONINFO(XrdHttpGetExtHandler, HttpTPC);
 
 
+void CurlDeleter::operator()(CURL *curl)
+{
+    if (curl) curl_easy_cleanup(curl);
+}
+
+
 // We need to utilize the full URL (including the query string), not just the
 // resource name.  The query portion is hidden in the `xrd-http-query` header;
 // we take this out and combine it with the resource name.
@@ -68,6 +75,56 @@ static std::string prepareURL(XrdHttpExtReq &req) {
 }
 
 
+//
+// When processing a redirection from the filesystem layer, it is permitted to return
+// some xrootd opaque data.  The quoting rules for xrootd opaque data are significantly
+// more permissive than a URI (basically, only '&' and '=' are disallowed while some
+// URI parsers may dislike characters like '"').  This function takes an opaque string
+// (e.g., foo=1&bar=2&baz=") and makes it safe for all URI parsers.
+std::string encode_xrootd_opaque_to_uri(CURL *curl, const std::string &opaque)
+{
+    std::stringstream parser(opaque);
+    std::string sequence;
+    std::stringstream output;
+    bool first = true;
+    while (getline(parser, sequence, '&')) {
+        if (sequence.empty()) {continue;}
+        size_t equal_pos = sequence.find('=');
+        char *val = NULL;
+        if (equal_pos != std::string::npos)
+            val = curl_easy_escape(curl, sequence.c_str() + equal_pos + 1, sequence.size()  - equal_pos - 1);
+        // Do not emit parameter if value exists and escaping failed.
+        if (!val && equal_pos != std::string::npos) {continue;}
+
+        if (!first) output << "&";
+        first = false;
+        output << sequence.substr(0, equal_pos);
+        if (val) {
+            output << "=" << val;
+            curl_free(val);
+        }
+    }
+    return output.str();
+}
+
+void
+TPCHandler::ConfigureCurlCA(CURL *curl)
+{
+    auto ca_filename = m_ca_file ? m_ca_file->CAFilename() : "";
+    auto crl_filename = m_ca_file ? m_ca_file->CRLFilename() : "";
+    if (!ca_filename.empty() && !crl_filename.empty()) {
+        curl_easy_setopt(curl, CURLOPT_CAINFO, ca_filename.c_str());
+        curl_easy_setopt(curl, CURLOPT_CRLFILE, crl_filename.c_str());
+    }
+    else if (!m_cadir.empty()) {
+        curl_easy_setopt(curl, CURLOPT_CAPATH, m_cadir.c_str());
+    }
+    if (!m_cafile.empty()) {
+        curl_easy_setopt(curl, CURLOPT_CAINFO, m_cafile.c_str());
+    }
+}
+
+
 bool TPCHandler::MatchesPath(const char *verb, const char *path) {
     return !strcmp(verb, "COPY") || !strcmp(verb, "OPTIONS");
 }
@@ -109,6 +166,8 @@ TPCHandler::~TPCHandler() {
 
 TPCHandler::TPCHandler(XrdSysError *log, const char *config, XrdOucEnv *myEnv) :
         m_desthttps(false),
+        m_timeout(60),
+        m_first_timeout(120),
         m_log(log->logger(), "TPC_"),
         m_sfs(NULL)
 {
@@ -137,7 +196,7 @@ std::string TPCHandler::GetAuthz(XrdHttpExtReq &req) {
     return authz;
 }
 
-int TPCHandler::RedirectTransfer(const std::string &redirect_resource,
+int TPCHandler::RedirectTransfer(CURL *curl, const std::string &redirect_resource,
     XrdHttpExtReq &req, XrdOucErrInfo &error, TPCLogRecord &rec)
 {
     int port;
@@ -163,7 +222,7 @@ int TPCHandler::RedirectTransfer(const std::string &redirect_resource,
     ss << "Location: http" << (m_desthttps ? "s" : "") << "://" << host << ":" << port << "/" << redirect_resource;
 
     if (!opaque.empty()) {
-      ss << "?" << opaque;
+      ss << "?" << encode_xrootd_opaque_to_uri(curl, opaque);
     }
 
     rec.status = 307;
@@ -219,14 +278,12 @@ int TPCHandler::DetermineXferSize(CURL *curl, XrdHttpExtReq &req, State &state,
         ss << "Remote server failed request: " << curl_easy_strerror(res);
         rec.status = 500;
         logTransferEvent(LogMask::Error, rec, "SIZE_FAIL", ss.str());
-        curl_easy_cleanup(curl);
         return req.SendSimpleResp(rec.status, NULL, NULL, const_cast<char *>(curl_easy_strerror(res)), 0);
     } else if (state.GetStatusCode() >= 400) {
         std::stringstream ss;
         ss << "Remote side failed with status code " << state.GetStatusCode();
         rec.status = 500;
         logTransferEvent(LogMask::Error, rec, "SIZE_FAIL", ss.str());
-        curl_easy_cleanup(curl);
         return req.SendSimpleResp(rec.status, NULL, NULL, const_cast<char *>(ss.str().c_str()), 0);
     } else if (res) {
         std::stringstream ss;
@@ -234,7 +291,6 @@ int TPCHandler::DetermineXferSize(CURL *curl, XrdHttpExtReq &req, State &state,
         rec.status = 500;
         logTransferEvent(LogMask::Error, rec, "SIZE_FAIL", ss.str());
         char msg[] = "Unknown internal transfer failure";
-        curl_easy_cleanup(curl);
         return req.SendSimpleResp(rec.status, NULL, NULL, msg, 0);
     }
     std::stringstream ss;
@@ -320,10 +376,11 @@ int TPCHandler::RunCurlWithUpdates(CURL *curl, XrdHttpExtReq &req, State &state,
         logTransferEvent(LogMask::Error, rec, "CURL_INIT_FAIL",
             "Failed to initialize a libcurl multi-handle");
         char msg[] = "Failed to initialize internal server memory";
-        curl_easy_cleanup(curl);
         return req.SendSimpleResp(rec.status, NULL, NULL, msg, 0);
     }
 
+    //curl_easy_setopt(curl, CURLOPT_BUFFERSIZE, 128*1024);
+
     CURLMcode mres;
     mres = curl_multi_add_handle(multi_handle, curl);
     if (mres) {
@@ -332,7 +389,6 @@ int TPCHandler::RunCurlWithUpdates(CURL *curl, XrdHttpExtReq &req, State &state,
         ss << "Failed to add transfer to libcurl multi-handle: " << curl_multi_strerror(mres);
         logTransferEvent(LogMask::Error, rec, "CURL_INIT_FAIL", ss.str());
         char msg[] = "Failed to initialize internal server handle";
-        curl_easy_cleanup(curl);
         curl_multi_cleanup(multi_handle);
         return req.SendSimpleResp(rec.status, NULL, NULL, msg, 0);
     }
@@ -340,7 +396,6 @@ int TPCHandler::RunCurlWithUpdates(CURL *curl, XrdHttpExtReq &req, State &state,
     // Start response to client prior to the first call to curl_multi_perform
     int retval = req.StartChunkedResp(201, "Created", "Content-Type: text/plain");
     if (retval) {
-        curl_easy_cleanup(curl);
         curl_multi_cleanup(multi_handle);
         logTransferEvent(LogMask::Error, rec, "RESPONSE_FAIL",
             "Failed to send the initial response to the TPC client");
@@ -354,19 +409,37 @@ int TPCHandler::RunCurlWithUpdates(CURL *curl, XrdHttpExtReq &req, State &state,
     // interrupt things to send back performance updates to the client.
     int running_handles = 1;
     time_t last_marker = 0;
+    // Track how long it's been since the last time we recorded more bytes being transferred.
+    off_t last_advance_bytes = 0;
+    time_t last_advance_time = time(NULL);
+    time_t transfer_start = last_advance_time;
     CURLcode res = static_cast<CURLcode>(-1);
     do {
         time_t now = time(NULL);
         time_t next_marker = last_marker + m_marker_period;
         if (now >= next_marker) {
+            off_t bytes_xfer = state.BytesTransferred();
+            if (bytes_xfer > last_advance_bytes) {
+                last_advance_bytes = bytes_xfer;
+                last_advance_time = now;
+            }
             if (SendPerfMarker(req, rec, state)) {
                 curl_multi_remove_handle(multi_handle, curl);
-                curl_easy_cleanup(curl);
                 curl_multi_cleanup(multi_handle);
                 logTransferEvent(LogMask::Error, rec, "PERFMARKER_FAIL",
                     "Failed to send a perf marker to the TPC client");
                 return -1;
             }
+            int timeout = (transfer_start == last_advance_time) ? m_first_timeout : m_timeout;
+            if (now > last_advance_time + timeout) {
+                state.SetErrorCode(10);
+                std::stringstream ss;
+                ss << "Transfer failed because no bytes have been received in " << timeout << " seconds.";
+                state.SetErrorMessage(ss.str());
+                curl_multi_remove_handle(multi_handle, curl);
+                curl_multi_cleanup(multi_handle);
+                break;
+            }
             last_marker = now;
         }
         mres = curl_multi_perform(multi_handle, &running_handles);
@@ -390,7 +463,6 @@ int TPCHandler::RunCurlWithUpdates(CURL *curl, XrdHttpExtReq &req, State &state,
                 CURL *easy_handle = msg->easy_handle;
                 res = msg->data.result;
                 curl_multi_remove_handle(multi_handle, easy_handle);
-                curl_easy_cleanup(easy_handle);
             }
         } while (msg);
 
@@ -416,7 +488,6 @@ int TPCHandler::RunCurlWithUpdates(CURL *curl, XrdHttpExtReq &req, State &state,
 
         char msg[] = "Internal server error due to libcurl";
         curl_multi_remove_handle(multi_handle, curl);
-        curl_easy_cleanup(curl);
         curl_multi_cleanup(multi_handle);
 
         if ((retval = req.ChunkResp(msg, 0))) {
@@ -436,13 +507,11 @@ int TPCHandler::RunCurlWithUpdates(CURL *curl, XrdHttpExtReq &req, State &state,
             CURL *easy_handle = msg->easy_handle;
             res = msg->data.result;
             curl_multi_remove_handle(multi_handle, easy_handle);
-            curl_easy_cleanup(easy_handle);
         }
     } while (msg);
 
-    if (res == -1) { // No transfers returned?!?
+    if (!state.GetErrorCode() && res == static_cast<CURLcode>(-1)) { // No transfers returned?!?
         curl_multi_remove_handle(multi_handle, curl);
-        curl_easy_cleanup(curl);
         curl_multi_cleanup(multi_handle);
         char msg[] = "Internal state error in libcurl";
         logTransferEvent(LogMask::Error, rec, "TRANSFER_CURL_ERROR", msg);
@@ -456,9 +525,17 @@ int TPCHandler::RunCurlWithUpdates(CURL *curl, XrdHttpExtReq &req, State &state,
     }
     curl_multi_cleanup(multi_handle);
 
+    state.Flush();
+
     rec.bytes_transferred = state.BytesTransferred();
     rec.tpc_status = state.GetStatusCode();
 
+    // Explicitly finalize the stream (which will close the underlying file
+    // handle) before the response is sent.  In some cases, subsequent HTTP
+    // requests can occur before the filesystem is done closing the handle -
+    // and those requests may occur against partial data.
+    state.Finalize();
+
     // Generate the final response back to the client.
     std::stringstream ss;
     bool success = false;
@@ -472,6 +549,14 @@ int TPCHandler::RunCurlWithUpdates(CURL *curl, XrdHttpExtReq &req, State &state,
         }
         logTransferEvent(LogMask::Error, rec, "TRANSFER_FAIL", ss2.str());
         ss << "failure: " << ss2.str();
+    } else if (state.GetErrorCode()) {
+        std::string err = state.GetErrorMessage();
+        if (err.empty()) {err = "(no error message provided)";}
+        else {std::replace(err.begin(), err.end(), '\n', ' ');}
+        std::stringstream ss2;
+        ss2 << "Error when interacting with local filesystem: " << err;
+        logTransferEvent(LogMask::Error, rec, "TRANSFER_FAIL", ss2.str());
+        ss << "failure: " << ss2.str();
     } else if (res != CURLE_OK) {
         std::stringstream ss2;
         ss2 << "HTTP library failure: " << curl_easy_strerror(res);
@@ -496,8 +581,17 @@ int TPCHandler::RunCurlBasic(CURL *curl, XrdHttpExtReq &req, State &state,
                              const char *log_prefix) {
     CURLcode res;
     res = curl_easy_perform(curl);
-    curl_easy_cleanup(curl);
-    if (res == CURLE_HTTP_RETURNED_ERROR) {
+    state.Flush();
+    state.Finalize();
+    if (state.GetErrorCode()) {
+        std::string err = state.GetErrorMessage();
+        if (err.empty()) {err = "(no error message provided)";}
+        else {std::replace(err.begin(), err.end(), '\n', ' ');}
+        std::stringstream ss2;
+        ss2 << "Error when interacting with local filesystem: " << err;
+        logTransferEvent(LogMask::Error, rec, "TRANSFER_FAIL", ss2.str());
+        ss << "failure: " << ss2.str();
+    } else if (res == CURLE_HTTP_RETURNED_ERROR) {
         m_log.Emsg(log_prefix, "Remote server failed request", curl_easy_strerror(res));
         return req.SendSimpleResp(500, NULL, NULL,
                                   const_cast<char *>(curl_easy_strerror(res)), 0);
@@ -526,14 +620,17 @@ int TPCHandler::ProcessPushReq(const std::string & resource, XrdHttpExtReq &req)
     char *name = req.GetSecEntity().name;
     if (name) rec.name = name;
     logTransferEvent(LogMask::Info, rec, "PUSH_START", "Starting a push request");
-    CURL *curl = curl_easy_init();
-    curl_easy_setopt(curl, CURLOPT_NOSIGNAL, 1);
+
+    ManagedCurlHandle curlPtr(curl_easy_init());
+    auto curl = curlPtr.get();
     if (!curl) {
         char msg[] = "Failed to initialize internal transfer resources";
         rec.status = 500;
         logTransferEvent(LogMask::Error, rec, "PUSH_FAIL", msg);
         return req.SendSimpleResp(rec.status, NULL, NULL, msg, 0);
     }
+    curl_easy_setopt(curl, CURLOPT_NOSIGNAL, 1);
+ 
     auto query_header = req.headers.find("xrd-http-fullresource");
     std::string redirect_resource = req.resource;
     if (query_header != req.headers.end()) {
@@ -545,7 +642,6 @@ int TPCHandler::ProcessPushReq(const std::string & resource, XrdHttpExtReq &req)
     AtomicEnd(m_monid_mutex);
     std::unique_ptr<XrdSfsFile> fh(m_sfs->newFile(name, file_monid));
     if (!fh.get()) {
-        curl_easy_cleanup(curl);
         rec.status = 500;
         logTransferEvent(LogMask::Error, rec, "OPEN_FAIL",
             "Failed to initialize internal transfer file handle");
@@ -559,10 +655,9 @@ int TPCHandler::ProcessPushReq(const std::string & resource, XrdHttpExtReq &req)
     int open_results = OpenWaitStall(*fh, full_url, SFS_O_RDONLY, 0644,
                                      req.GetSecEntity(), authz);
     if (SFS_REDIRECT == open_results) {
-        curl_easy_cleanup(curl);
-        return RedirectTransfer(redirect_resource, req, fh->error, rec);
+        int result = RedirectTransfer(curl, redirect_resource, req, fh->error, rec);
+        return result;
     } else if (SFS_OK != open_results) {
-        curl_easy_cleanup(curl);
         int code;
         char msg_generic[] = "Failed to open local resource";
         const char *msg = fh->error.getErrText(code);
@@ -576,9 +671,7 @@ int TPCHandler::ProcessPushReq(const std::string & resource, XrdHttpExtReq &req)
         fh->close();
         return resp_result;
     }
-    if (!m_cadir.empty()) {
-            curl_easy_setopt(curl, CURLOPT_CAPATH, m_cadir.c_str());
-    }
+    ConfigureCurlCA(curl);
     curl_easy_setopt(curl, CURLOPT_URL, resource.c_str());
 
     Stream stream(std::move(fh), 0, 0, m_log);
@@ -600,17 +693,18 @@ int TPCHandler::ProcessPullReq(const std::string &resource, XrdHttpExtReq &req)
     char *name = req.GetSecEntity().name;
     if (name) rec.name = name;
     logTransferEvent(LogMask::Info, rec, "PULL_START", "Starting a push request");
-    CURL *curl = curl_easy_init();
-    curl_easy_setopt(curl, CURLOPT_NOSIGNAL, 1);
+
+    ManagedCurlHandle curlPtr(curl_easy_init());
+    auto curl = curlPtr.get();
     if (!curl) {
             char msg[] = "Failed to initialize internal transfer resources";
             rec.status = 500;
             logTransferEvent(LogMask::Error, rec, "PULL_FAIL", msg);
             return req.SendSimpleResp(rec.status, NULL, NULL, msg, 0);
     }
+    curl_easy_setopt(curl, CURLOPT_NOSIGNAL, 1);
     std::unique_ptr<XrdSfsFile> fh(m_sfs->newFile(name, m_monid++));
     if (!fh.get()) {
-            curl_easy_cleanup(curl);
             char msg[] = "Failed to initialize internal transfer file handle";
              rec.status = 500;
             logTransferEvent(LogMask::Error, rec, "PULL_FAIL", msg);
@@ -636,13 +730,12 @@ int TPCHandler::ProcessPullReq(const std::string &resource, XrdHttpExtReq &req)
             } catch (...) { // Handled below
             }
             if (stream_req < 0 || stream_req > 100) {
-                curl_easy_cleanup(curl);
                 char msg[] = "Invalid request for number of streams";
                 rec.status = 500;
                 logTransferEvent(LogMask::Info, rec, "INVALID_REQUEST", msg);
                 return req.SendSimpleResp(rec.status, NULL, NULL, msg, 0);
             }
-            streams = streams == 0 ? 1 : stream_req;
+            streams = stream_req == 0 ? 1 : stream_req;
         }
     }
     rec.streams = streams;
@@ -652,10 +745,9 @@ int TPCHandler::ProcessPullReq(const std::string &resource, XrdHttpExtReq &req)
     int open_result = OpenWaitStall(*fh, full_url, mode|SFS_O_WRONLY, 0644,
                                     req.GetSecEntity(), authz);
     if (SFS_REDIRECT == open_result) {
-        curl_easy_cleanup(curl);
-        return RedirectTransfer(redirect_resource, req, fh->error, rec);
+        int result = RedirectTransfer(curl, redirect_resource, req, fh->error, rec);
+        return result;
     } else if (SFS_OK != open_result) {
-        curl_easy_cleanup(curl);
         int code;
         char msg_generic[] = "Failed to open local resource";
         const char *msg = fh->error.getErrText(code);
@@ -669,11 +761,9 @@ int TPCHandler::ProcessPullReq(const std::string &resource, XrdHttpExtReq &req)
         fh->close();
         return resp_result;
     }
-    if (!m_cadir.empty()) {
-        curl_easy_setopt(curl, CURLOPT_CAPATH, m_cadir.c_str());
-    }
+    ConfigureCurlCA(curl);
     curl_easy_setopt(curl, CURLOPT_URL, resource.c_str());
-    Stream stream(std::move(fh), streams * m_pipelining_multiplier, m_block_size, m_log);
+    Stream stream(std::move(fh), streams * m_pipelining_multiplier, streams > 1 ? m_block_size : m_small_block_size, m_log);
     State state(0, stream, curl, false);
     state.CopyHeaders(req);
 
